
/* 
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Id: espda_wrap.c 161538 2014-05-22 13:56:10Z gcupani $
 * $Author: gcupani $
 * $Date: 2014-05-22 15:56:10 +0200 (Thu, 22 May 2014) $
 * $Revision: 161538 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_wrap.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup   espda_wrap  Utilities for wrapping
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/

cpl_error_code espda_wrap_check(cpl_frameset *set, cpl_recipeconfig *cfg,
    const char *cfg_tag) {
    //extern int v_level;
    espda_util_null(set);
    espda_util_null(cfg);
    espda_check_name(set, cfg, cfg_tag);
    espda_check_tag (set, cfg, cfg_tag);
    espda_check_num (set, cfg, cfg_tag);
    return cpl_error_get_code();
}


/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/

cpl_error_code espda_wrap_line(cpl_table  *fspec_in,
                               cpl_table  *rspec_in,
                               cpl_table  *lfit_in,
                               cpl_table  *ion,
                               cpl_table  *voigt_func,
                               double      model_resol,
                               int         iter_max,
                               int         grp_sel,
                               int         conv_flag,
                               cpl_table  *fprof,
                               cpl_table  *rprof,
                               cpl_table **fspec_out,
                               cpl_table **rspec_out,
                               cpl_table **lfit_out,
                               cpl_array **rchisq,
                               cpl_array **iter)

{
    //cpl_errorstate prestate = cpl_errorstate_get();
    extern int v_level;
    extern int   cycle;
    extern char *grp_flag, *cycle_flag;
        
    if (lfit_in != NULL) {

        // Select the group of lines to be fitted (grp_sel = 0 means all groups)
        cpl_array *grp_list = NULL;
        cpl_table *fspec_sel = NULL;
        cpl_table *rspec_sel = NULL;
        cpl_table *lfit_sel = NULL;
        espda_util_func(espda_split_group(fspec_in, rspec_in, lfit_in, grp_sel,
					  &grp_list, &fspec_sel, &rspec_sel,
					  &lfit_sel), SPLIT_GROUP, 0);

        // Create and interpolate the instrument profile
        cpl_table *inst_prof = NULL;
        if (conv_flag == 1) {
            if (fprof == NULL || rprof == NULL) {
                espda_print_verb(1, cpl_msg_info(cpl_func, "Creating and "
                    "interpolating instrument profile..."));
                inst_prof = espda_model_inst(MODEL_SIZE, model_resol,
					     MODEL_KAPPA);
            }
            if (fspec_sel != NULL && fprof == NULL) {
                cpl_ensure_code(espda_interp_inst(fspec_sel, inst_prof,
						  MODEL_SIZE, model_resol,
						  MODEL_KAPPA, &fprof) 
                                == CPL_ERROR_NONE, cpl_error_get_code());
            }
	    if (rprof == NULL) {
	        cpl_ensure_code(espda_interp_inst(rspec_sel, inst_prof, 
						  MODEL_SIZE, model_resol,
						  MODEL_KAPPA, &rprof) 
                            == CPL_ERROR_NONE, cpl_error_get_code());
            }
        }
        cpl_table_delete(inst_prof);

        cpl_table *lfit_null = NULL;
        *rchisq = cpl_array_new(4, CPL_TYPE_DOUBLE);
        *iter   = cpl_array_new(2, CPL_TYPE_INT);

        // Loop over groups
        int grp_max  = (grp_sel > 0) ? (int)fmin(cpl_array_get_size(grp_list), 
                            grp_sel) : cpl_array_get_size(grp_list);
        int grp_min  = (grp_sel <= grp_max) ? (int)fmax(0, grp_sel - 1) : 0;
        //char *part_name  = "chunk:   ";

	//int grp;
	cpl_table *lfit_chunk = NULL;
	cpl_size lfit_size;
	cpl_table *fspec_over  = NULL, *fspec_chunk = NULL;
	cpl_table *fprof_chunk = NULL;
	cpl_table *rspec_over  = NULL, *rspec_chunk = NULL;
	cpl_table *rprof_chunk = NULL;
	/*
#pragma omp parallel private(grp, lfit_chunk, lfit_size, fspec_over,	\
			     fspec_chunk, fprof_chunk, rspec_over, rspec_chunk,\
			     rprof_chunk)
	*/
	{
	  //#pragma omp for
        for (int g = grp_min; g <= grp_max; g++) {
	    int grp = 0;
	    //grp = 0;
            //cpl_table *lfit_chunk = NULL;
            grp_flag   = "NULL";
            if (g < grp_max) {
                if (g == grp_min) {
                    grp_flag = "MIN";
                }
                grp       = cpl_array_get_int(grp_list, g, NULL); 
                cpl_table_unselect_all(lfit_sel);
                cpl_table_or_selected_int(lfit_sel, "GRPID", CPL_EQUAL_TO, grp);
                lfit_chunk = cpl_table_extract_selected(lfit_sel);
            }
            else {
	        grp_flag = "MAX";
                //part_name = "spectrum:";
                lfit_chunk = cpl_table_duplicate(*lfit_out);
            }
            //cpl_size   lfit_size = cpl_table_get_nrow(lfit_chunk);
	    lfit_size = cpl_table_get_nrow(lfit_chunk);
	    //printf("%i %i\n", g, lfit_size);
            //cpl_table *fspec_over  = NULL, *fspec_chunk = NULL;
            //cpl_table *fprof_chunk = NULL;
            if (fspec_sel != NULL) {
	      /*
		espda_util_func(espda_split_chunk(fspec_sel, fprof, 
                    lfit_chunk, grp_flag, &fspec_over, &fspec_chunk,
                    &fprof_chunk), SPLIT_CHUNK, 0);
	      */
	      espda_split_chunk(fspec_sel, fprof, 
                    lfit_chunk, grp_flag, &fspec_over, &fspec_chunk,
				&fprof_chunk);
            }
            //cpl_table *rspec_over  = NULL, *rspec_chunk = NULL;
            //cpl_table *rprof_chunk = NULL;
            /*
	      espda_util_func(espda_split_chunk(rspec_sel, rprof, 
                lfit_chunk, grp_flag, &rspec_over, &rspec_chunk, 
                &rprof_chunk), SPLIT_CHUNK, 0);
	    */
	    espda_split_chunk(rspec_sel, rprof, 
                lfit_chunk, grp_flag, &rspec_over, &rspec_chunk, 
			      &rprof_chunk);
	    //printf("%i %i %i\n", cpl_table_get_nrow(rspec_sel),
	    //	   cpl_table_get_nrow(rspec_over),
	    //	   cpl_table_get_nrow(rspec_chunk));
            if (g < grp_max) {
                espda_print_verb(1, cpl_msg_info(cpl_func, "Group %i; "
                    "wavelength ranges:", grp + 1));
                for (int i = 0; i < lfit_size; i++) {
                    espda_print_verb(1, cpl_msg_info(cpl_func, " %2i. "
                        "[%6.2f,%6.2f]", i + 1,
                        cpl_table_get_double(lfit_chunk, "WAVELMIN", i, NULL),
                        cpl_table_get_double(lfit_chunk, "WAVELMAX", i, NULL)));
                }
            }
            else {
                espda_print_verb(1, cpl_msg_info(cpl_func, "Overall fit:"));
            }

            cycle = 0;
            cycle_flag = "rebinned ";
            /*
	      espda_util_func(espda_fit_voigt(rspec_over, rspec_chunk,
                lfit_chunk, rprof_chunk, ion, voigt_func, *rchisq, *iter, 
                iter_max, rspec_out, lfit_out), FIT_VOIGT, 0);
	    */
	    espda_fit_voigt(rspec_over, rspec_chunk,
                lfit_chunk, rprof_chunk, ion, voigt_func, *rchisq, *iter, 
			    iter_max, rspec_out, lfit_out);
            if (fspec_over != NULL && fspec_chunk != NULL) {
                cycle = 1;
                cycle_flag = "equalized";
                /*
		  espda_util_func(espda_fit_voigt(fspec_over, fspec_chunk,    
                    lfit_chunk, fprof_chunk, ion, voigt_func, *rchisq, *iter, 
                    iter_max, fspec_out, &lfit_null), FIT_VOIGT, 0);
		*/
		espda_fit_voigt(fspec_over, fspec_chunk,    
                    lfit_chunk, fprof_chunk, ion, voigt_func, *rchisq, *iter, 
				iter_max, fspec_out, &lfit_null);
		cpl_array_dump(*rchisq, 0, 10, NULL);
            }
	    /*
            cpl_table_delete(fspec_over);
            cpl_table_delete(fspec_chunk);
            cpl_table_delete(fprof_chunk);
            cpl_table_delete(rspec_over);
            cpl_table_delete(rspec_chunk);
            cpl_table_delete(rprof_chunk);
            cpl_table_delete(lfit_chunk);
	    */
        }
	}
	cpl_table_delete(fspec_over);
	cpl_table_delete(fspec_chunk);
	cpl_table_delete(fprof_chunk);
	cpl_table_delete(rspec_over);
	cpl_table_delete(rspec_chunk);
	cpl_table_delete(rprof_chunk);
	cpl_table_delete(lfit_chunk);
        espda_print_verb(2, espda_print_line(*lfit_out));

       

        cpl_table_delete(fspec_sel);
        cpl_table_delete(rspec_sel);
        cpl_table_delete(lfit_sel);
        //cpl_table_delete(fprof);
        //cpl_table_delete(rprof);
        cpl_array_delete(grp_list);

    }
    else {
        grp_flag   = "MAX";

        // Create and interpolate the instrument profile
        cpl_table *inst_prof = NULL;
        cpl_table *fprof    = NULL;
        cpl_table *rprof = NULL;
        if (conv_flag == 1) {
            espda_print_verb(1, cpl_msg_info(cpl_func, "Creating and "
                "interpolating instrument profile..."));
            inst_prof = espda_model_inst(MODEL_SIZE, model_resol, MODEL_KAPPA);
            if (fspec_in != NULL) {
                cpl_ensure_code(espda_interp_inst(fspec_in, inst_prof, MODEL_SIZE,
                                model_resol, MODEL_KAPPA, &fprof) 
                                == CPL_ERROR_NONE, cpl_error_get_code());
            }
            cpl_ensure_code(espda_interp_inst(rspec_in, inst_prof, MODEL_SIZE, 
                            model_resol, MODEL_KAPPA, &rprof) 
                            == CPL_ERROR_NONE, cpl_error_get_code());
        }
        cpl_table_delete(inst_prof);

        cpl_table *fspec_over  = NULL, *fspec_chunk = NULL;
        cpl_table *fprof_chunk = NULL;
        if (fspec_in != NULL) {
            espda_util_func(espda_split_chunk(fspec_in, fprof, 
                NULL, grp_flag, &fspec_over, &fspec_chunk,
                &fprof_chunk), SPLIT_CHUNK, 0);
        }
        cpl_table *rspec_over  = NULL, *rspec_chunk = NULL;
        cpl_table *rprof_chunk = NULL;

        espda_util_func(espda_split_chunk(rspec_in, rprof, 
            NULL, grp_flag, &rspec_over, &rspec_chunk, 
            &rprof_chunk), SPLIT_CHUNK, 0);
        espda_struct_check(fspec_chunk,    *fspec_out = 
            cpl_table_duplicate(fspec_chunk));
        espda_struct_check(rspec_chunk, *rspec_out =
            cpl_table_duplicate(rspec_chunk));
        if (*fspec_out    != NULL) {
            cpl_table_fill_column_window_double(*fspec_out,    "FIT", 0, 
                cpl_table_get_nrow(*fspec_out),    1.0);
        }
        if (*rspec_out != NULL) {
            cpl_table_fill_column_window_double(*rspec_out, "FIT", 0, 
                cpl_table_get_nrow(*rspec_out), 1.0);
        }        
    }
    /*
    if (!cpl_errorstate_is_equal(prestate)) {
        cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to fit Voigt profiles to line list.");
	       }*/
    return cpl_error_get_code();
}

/**@}*/
